/* Copyright (C) 2016 Centro de Computacao Cientifica e Software Livre
 * Departamento de Informatica - Universidade Federal do Parana - C3SL/UFPR
 *
 * This file is part of simmc-agent
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
 * USA.
 */

#include <libudev.h>
#include <fstream>
#include <string>
#include <iostream>
#include <list>
#include "agent/linux/get_disks_info.h"


void get_scsi_disks(std::list<disk_t>& disks) {  // NOLINT(runtime/references)

    struct udev *udev;
    struct udev_enumerate *enumerate;
    struct udev_list_entry *devices, *dev_list_entry, *links_list_entry;
    struct udev_device *dev;

    udev = udev_new();
    if (!udev) {
        throw "Could not creat udev\n";
    }

    /* SCSI disk devices MAJOR = {8, 65-71, 128-135}
    see https://www.kernel.org/doc/Documentation/devices.txt */
    const char* scsi_majors[] = {"8", "65", "66", "67", "68", "69", "70", "71",
                                 "128", "129", "130", "131", "132", "133",
                                 "134", "135"};
    int i;

    for (i=0; i < 16; i++) {

        /* create a list of the disk devices in the 'block' subsystem and with
        MAJOR == one of the possible scsi major numbers. */
        enumerate = udev_enumerate_new(udev);
        udev_enumerate_add_match_property(enumerate, "MAJOR", scsi_majors[i]);
        udev_enumerate_add_match_subsystem(enumerate, "block");
        udev_enumerate_scan_devices(enumerate);
        devices = udev_enumerate_get_list_entry(enumerate);

        udev_list_entry_foreach(dev_list_entry, devices) {
            /* potential new item */
            disk_t curr = {.model = "", .size = -1};

            /* get the filename of the /sys entry for the device
            and create udev_device object representing it */
            const char *path;
            path = udev_list_entry_get_name(dev_list_entry);
            dev = udev_device_new_from_syspath(udev, path);

            /* MINOR = (16 * drive_number) + partition_number
            partition_number is 0 for the whole drive, which is the one we want
            therefore, if mod 16 != 0, skip device */
            const char* minor = udev_device_get_property_value(dev, "MINOR");
            if ((atoi(minor) % 16) != 0) {
                udev_device_unref(dev);
                continue;
            }


            const char* serial = udev_device_get_property_value(dev,
                                                       "ID_SERIAL");
            const char* size = udev_device_get_sysattr_value(dev, "size");

            if (serial) {
                curr.model.assign(std::string(serial));
            }

            /* size in Bytes = size * 512
            size in GB = (size * 2^9)/2^30 = size/2^21 */
            if (size) {
                curr.size = atoll(size) >> 21;
            }

            /* do not push some anomalous device whose model and size we
            couldn't obtain */
            if (size && serial) {
                disks.push_front(curr);
            }

            udev_device_unref(dev);
        }

        /* free the enumerator object */
        udev_enumerate_unref(enumerate);

    }

    udev_unref(udev);
}


void get_disks_info(std::list<disk_t>& disks) {  // NOLINT(runtime/references)
    get_scsi_disks(disks);

    if (disks.empty()) {
        throw std::string("No disk info found");
    }
}
